#include <QtGui/QMouseEvent>
#include <cassert>
#include <cmath>
#include <GL/glu.h>
#include "GLWidget.h"
#include "R3Graph.h"
#include "icosahed.h"

static void vertex(const R3Point& p);

static const int NUM_COLORS = 8;
static const GLfloat COLORS[NUM_COLORS][3] = {
    { 1.0, 0.0, 0.0 },      // red
    { 0.0, 1.0, 0.0 },      // green
    { 0.0, 0.0, 1.0 },      // blue
    { 1.0, 1.0, 1.0 },      // white
    { 1.0, 1.0, 0.0 },      // yellow
    { 0.0, 1.0, 1.0 },      // cyan
    { 1.0, 0.0, 1.0 },      // magenta
    { 0.2, 0.2, 0.2 }       // dark gray
};

GLWidget::GLWidget(QWidget *parent):
    QGLWidget(parent),
    icosahedron(),
    wireframe(true),
    useColors(true),
    xmin(-2. * 4./3.),
    xmax(2. *4./3.),
    ymin(-2.),
    ymax(2.),
    alpha(ALPHA0),
    beta(BETA0),
    mouseX(-1),     // Undefined
    mouseY(-1)      // Undefined
{
    icosahedron.initialize();
    setMouseTracking(true);
}

void GLWidget::initializeGL() {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glDepthFunc(GL_LEQUAL);

    // Lighting
    glEnable(GL_LIGHT0);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

    GLfloat pos[4];
    pos[0] = 0.; pos[1] = 0.; pos[2] = 1.; pos[3] = 0.; // Directional light
    glLightfv(GL_LIGHT0, GL_POSITION, pos);

    GLfloat light[4];
    light[0] = 0.25; light[1] = 0.25; light[2] = 0.25; light[3] = 1.;
    glLightfv(GL_LIGHT0, GL_AMBIENT, light);

    light[0] = 1.; light[1] = 1.; light[2] = 1.; light[3] = 1.;
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void GLWidget::resizeGL(int w, int h) {
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (w == 0)
        w = 1;
    if (h == 0)
        h = 1;
    double aspect = (double) w / (double) h;
    assert(aspect > 0.);

    ymin = xmin / aspect;
    ymax = (-ymin);
    double depth = ymax - ymin;
    glOrtho(
        xmin, xmax,
        ymin, ymax,
        -10. * depth, 10. * depth
    );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void GLWidget::paintGL() {
    // Rotate the scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_LIGHT0);        // Turn on the first light source

    // Rotate to the cabinet projection
    //... glRotatef(30., 1., 0., 0.);
    //... glRotatef(-30., 0., 1., 0.);

    // Go to mathematical coord. system: Z-axis goes up
    //... glRotatef(-90., 1., 0., 0.);

    // Rotate a model
    glRotatef(beta, 1., 0., 0.);
    glRotatef(alpha, 0., 1., 0.);

    render();           // Draw a scene graph
}

void GLWidget::mousePressEvent(QMouseEvent * /* event */) {
    mouseX = (-1);      // Undefined
    mouseY = (-1);      // Undefined
}

void GLWidget::mouseReleaseEvent(QMouseEvent * /* event */) {
    mouseX = (-1);      // Undefined
    mouseY = (-1);      // Undefined
}

void GLWidget::mouseMoveEvent(QMouseEvent *event) {
    //... printf("%d, %d\n", event->x(), event->y());
    int x = event->x();
    int y = event->y();
    if ((event->buttons() & Qt::LeftButton) != 0) {
        if (mouseX >= 0) {
            int dx = x - mouseX;
            int dy = y - mouseY;
            if (beta > 100. || beta < (-100.))
                dx = -dx;

            alpha += dx * 0.1;
            if (fabs(alpha) > 360.)
                alpha -= ((int) alpha / 360) * 360.;

            beta += dy * 0.1;
            if (fabs(beta) > 360.)
                beta -= ((int) beta / 360) * 360.;

            repaint();
        }
        mouseX = x; mouseY = y;

    } else {
        mouseX = (-1);      // Undefined
        mouseY = (-1);      // Undefined
    }
}

void GLWidget::keyPressEvent(QKeyEvent* event) {
    switch (event->key()) {
    case Qt::Key_S:     // s => split faces
        if (icosahedron.numSplits < 7)
            icosahedron.split();
        else
            icosahedron.initialize();
        repaint();
        break;
    case Qt::Key_W:     // w => wireframe off/on
        wireframe = !wireframe;
        repaint();
        repaint();  //???
        break;
    case Qt::Key_C:     // w => use colors off/on
        useColors = !useColors;
        repaint();
        repaint();  //???
        break;
    case Qt::Key_I:     // i => initialize
        icosahedron.initialize();
        repaint();
        break;
    case Qt::Key_Escape:
    case Qt::Key_Q:
        close();
        break;
    case Qt::Key_Space:
        alpha = ALPHA0;
        beta = BETA0;
        repaint();
        break;
    default:
        event->ignore();
        break;
    }
}

//
// Draw a 3D model
//
void GLWidget::render() {
    glClearColor(0.2, 0.3, 0.5, 1.); // Background color: dark blue
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    GLfloat color[4];
    color[3] = 1.;
    if (wireframe) {
        color[0] = 0.7;
        color[1] = 0.7;
        color[2] = 0.7;
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
        glBegin(GL_LINES);
    } else {
        glBegin(GL_TRIANGLES);
    }

    for (int i = 0; i < icosahedron.numFaces; ++i) {
        if (!wireframe) {
            R3Vector n = (
                icosahedron.vertices[icosahedron.faces[i][1]] -
                icosahedron.vertices[icosahedron.faces[i][0]]
            ).vectorProduct(
                icosahedron.vertices[icosahedron.faces[i][2]] -
                icosahedron.vertices[icosahedron.faces[i][0]]
            );
            n.normalize();
            glNormal3d(n.x, n.y, n.z);

            if (useColors) {
                const GLfloat *c = COLORS[i % NUM_COLORS];

                color[0] = c[0];
                color[1] = c[1];
                color[2] = c[2];
            } else {
                // Silver color
                color[0] = 0.6;
                color[1] = 0.7;
                color[2] = 0.8;
            }
            glMaterialfv(
                GL_FRONT_AND_BACK,
                GL_AMBIENT_AND_DIFFUSE,
                color
            );
        }

        for (int j = 0; j < 3; ++j) {
            if (wireframe) {
                R3Point& v0 = icosahedron.vertices[
                    icosahedron.faces[i][j]
                ];
                R3Point& v1 = icosahedron.vertices[
                    icosahedron.faces[i][(j+1)%3]
                ];
                R3Vector n = (
                    (v0-icosahedron.center()) +
                    (v1-icosahedron.center())
                ).normalize();
                glNormal3f(n.x, n.y, n.z);
                vertex(v0);
                vertex(v1);
            } else {
                vertex(
                    icosahedron.vertices[
                        icosahedron.faces[i][j]
                    ]
                );
            }
        }
    }
    glEnd();
}

static void vertex(const R3Point& p) {
    glVertex3d(
        (GLfloat) p.x, (GLfloat) p.y, (GLfloat) p.z
    );
}
